<?php
/**
 * @file
 * Provides page callbacks and their helper functions.
 */

/**
 * Page callback; Function that returns the block content for a Twitter feed.
 *
 * @param $config
 *   The block configuration array.
 * @param $page
 *   Page to load.
 * @param $type
 *   Type of callback, ajax or nojs.
 *
 * @return
 *   Render array for the block.
 */
function selective_tweets_twitter_feed_build($config, $page = 0, $type = 'nojs') {
  $data = unserialize($config['data']);
  $config['data'] = $data;
  $bid = $config['bid'];

  // When first loaded, take initial_count as number of tweets to load.
  $count = ($type == 'ajax') ? $data['count'] : $data['initial_count'];

  // Retrieve latest tweet ids from cache.
  try {
    $display_mode = $data['display_mode'];
    $tweet_ids = selective_tweets_twitter_feed_get($config, $display_mode, $page);
  } catch (Exception $ex) {
    // Serve fault message when something went wrong.
    unset($_SESSION['messages']);
    drupal_set_message($ex->getMessage(), 'error');
    $render['tweet_list']['#markup'] = theme('status_messages');

    return $render;
  }

  // Render the tweets.
  $render['tweet_list']['#markup'] = selective_tweets_feed_render($tweet_ids, $config);

  // Determine whether a load more link is needed or not.
  if (count($tweet_ids) == $count && $data['count'] != 0) {
    $render['load_more'] = array(
      '#theme' => 'link',
      '#text' => t('Load more'),
      '#path' => 'selective-tweets/feed/load/' . $bid . '/' . (++$page) . '/nojs',
      '#options' => array(
        'html' => TRUE,
        'attributes' => array(
          'id' => 'load-more-tweets-bid-' . $bid,
          'class' => array('use-ajax', 'load-more'),
        ),
      ),
    );
  }

  // If ajax request, insert the ajax commands and stop the page rendering.
  if ($type == 'ajax') {
    $commands[] = ajax_command_invoke('#block-selective-tweets-' . $bid . ' .selective-tweet:last', 'after', array(drupal_render($render['tweet_list'])));
    if (!empty($render['load_more'])) {
      $commands[] = ajax_command_replace('#load-more-tweets-bid-' . $bid, drupal_render($render['load_more']));
    }
    else {
      $commands[] = ajax_command_remove('#load-more-tweets-bid-' . $bid);
    }

    // Render as ajax response.
    ajax_deliver(array(
      '#type' => 'ajax',
      '#commands' => $commands,
    ));

    drupal_exit();
  }

  return $render;
}

/**
 * Helper function to render a twitter feed based on tweet IDs.
 *
 * @param $ids
 *   An array of tweet IDs to be rendered.
 * @param $block_config
 *   The configuration array of the block.
 *
 * @return
 *   A string containing the rendered tweets (HTML).
 */
function selective_tweets_feed_render($ids, $block_config) {
  $output = '';

  try {
    foreach ($ids as $tid) {
      $output .= selective_tweets_tweet_render($tid, $block_config);
    }
  } catch (Exception $ex) {
    drupal_set_message(t("Sorry, we're not able to display Tweets at this moment. Try again later."), 'error');
    watchdog('selective_tweets', $ex->getMessage(), array(), WATCHDOG_ERROR);
  }

  return $output;
}

/**
 * Helper function to convert one tweet ID to HTML.
 *
 * @param $tid
 *   The tweet ID.
 * @param $block_config
 *   The configuration array of the block.
 *
 * @return
 *   A tweet as HTML.
 */
function selective_tweets_tweet_render($tid, $block_config) {
  // We separate the tweets per block in the cache table, so
  // they can be cleared separately.
  $cid = 'selective_tweets::bid_' . $block_config['bid'] . '_rendered_tweets';
  $cache = cache_get($cid);

  // Check if rendered Tweet is cached already.
  if ($cache && isset($cache->data[$tid])) {
    return $cache->data[$tid];
  }
  else {
    try {
      // Get the current cache to add the new tweet to it.
      $updated_cache = array();
      if (is_object($cache) && isset($cache->data)) {
        $updated_cache = $cache->data;
      }

      // Authentication Twitter API.
      $connection = selective_tweets_twitter_connect();
      $api_version = variable_get('selective_tweets_api_version', SELECTIVE_TWEETS_API_VERSION);

      switch ($block_config['data']['render_mode']) {
        case SELECTIVE_TWEETS_EMBED:
          $params = array(
            'id' => $tid,
            'hide_media' => $block_config['data']['hide_media'],
            'hide_thread' => $block_config['data']['hide_thread'],
            'maxwidth' => $block_config['data']['maxwidth'],
            'omit_script' => TRUE,
          );

          // Convert a tweet ID to an embedded Tweet via the Twitter API.
          $response = json_decode(
            $connection->oAuthRequest(
              'https://api.twitter.com/' . $api_version . '/statuses/oembed.json',
              'GET',
              $params
            )
          );

          $updated_cache[$tid] = '<div class="selective-tweet tweet-id-' . $tid . '">' . $response->html . '</div>';
          break;

        case SELECTIVE_TWEETS_RAW:
          $params = array(
            'id' => $tid,
            'trim_user' => FALSE,
            'include_my_retweet' => FALSE,
            'include_entities' => TRUE,
          );

          // Convert a tweet ID to an embedded Tweet via the Twitter API.
          $response = json_decode(
            $connection->oAuthRequest(
              'https://api.twitter.com/' . $api_version . '/statuses/show.json',
              'GET',
              $params
            )
          );

          if ($block_config['data']['insert_links']) {
            selective_tweets_insert_links($response);
          }

          // Let other modules alter the tweet object.
          drupal_alter('selective_tweet_raw', $response, $block_config['bid']);

          $updated_cache[$tid] = theme('selective_tweet_raw', array('tweet' => $response));
          break;
      }

      // Set cache.
      cache_set($cid, $updated_cache, 'cache', CACHE_PERMANENT);

      return $updated_cache[$tid];

    } catch (Exception $ex) {
      watchdog('selective_tweets', $ex->getMessage(), array(), WATCHDOG_ERROR);
    }
  }

  return '';
}

/**
 * Helper function to get the tweets from tweet stack in cache. If the caching
 * time exceeded, the cache will be supplemented with new tweets.
 *
 * @param $block_config
 *   The configuration array of the block.
 * @param $display_mode
 *   The type of feed to load: live or manual.
 * @param $page
 *   The page to load.
 *
 * @return
 *   An array containing the tweet IDs.
 */
function selective_tweets_twitter_feed_get($block_config, $display_mode = NULL, $page = 0) {
  // If caching time period exceeded, fill cache with new tweets, but keep the
  // old ones as they can still be displayed.
  $cid = 'selective_tweets::bid_' . $block_config['bid'] . '_tweet_stack';
  $cache_expire = $block_config['cache_expire'];

  $cache = cache_get($cid);

  // Only get new tweets from API when first page is loaded AND when no cache is
  // available or the Selective Tweets cache for this block is expired.
  if ((REQUEST_TIME >= $block_config['cache_expire'] || !$cache) && $page == 0) {
    // Add newest Tweets to the tweet stack in cache.
    $tweet_ids = selective_tweets_build_cache($cid, $cache, $block_config, $display_mode);
  }
  else {
    $tweet_ids = $cache->data;
  }

  // Get the tweets of the requested page out of the array.
  // These calculations are made because the initial number of tweets can differ
  // from the ones that get loaden on the next pages.
  if ($page > 0) {
    $offset = $block_config['data']['initial_count'] + (--$page * $block_config['data']['count']);
    $count = $block_config['data']['count'];
  }
  else {
    $offset = 0;
    $count = $block_config['data']['initial_count'];
  }

  return array_slice($tweet_ids, $offset, $count);
}

/**
 * Helper function to get tweets from the Twitter API and cache them in order
 * to save expensive calls to Twitter.
 *
 * @param $cid
 *   The cache ID where the tweets for this block are cached.
 * @param $cache
 *   The currently cached tweet IDs.
 * @param $block_config
 *   The configuration array of the block.
 * @param $display_mode
 *   The mode which tweets to display.
 *
 * @return
 *   An array containing all cached tweet IDs.
 */
function selective_tweets_build_cache($cid, $cache, $block_config, $display_mode) {
  // Get currently cached tweets and the last cached tweet id.
  $cached_tweet_ids = array();
  $since_id = NULL;
  if ($cache) {
    $cached_tweet_ids = $cache->data;
    $since_id = reset($cached_tweet_ids);
  }

  $api_version = variable_get('selective_tweets_api_version', SELECTIVE_TWEETS_API_VERSION);
  $data = $block_config['data'];

  // Authentication Twitter API.
  $connection = selective_tweets_twitter_connect();

  // Default params for all requests.
  $params = array(
    'screen_name' => $data['screen_name'],
    'trim_user' => TRUE,
    'exclude_replies' => TRUE,
    'include_rts' => $data['include_retweets'],
    'include_entities' => FALSE,
    'result_type' => 'recent',
  );

  switch ($display_mode) {
    case ST_SELECTION_MODE_SEARCH:
      $params += array(
        'count' => 200,
        'q' => '',
      );
      if ($data['from_account'] != '') {
        $params['q'] .= ($params['q'] == '') ? 'from:' : ' from:';
        $data['from_account'] = str_replace(' ', ' OR from:', $data['from_account']);
        $params['q'] .= $data['from_account'];
      }
      if ($data['to_account'] != '') {
        $params['q'] .= ($params['q'] == '') ? 'to:' : ' to:';
        $data['to_account'] = str_replace(' AND ', ' AND to:', $data['to_account']);
        $data['to_account'] = str_replace(' OR ', ' OR to:', $data['to_account']);
        $params['q'] .= $data['to_account'];
      }
      if ($data['user_mentions'] != '') {
        $params['q'] .= ($params['q'] == '') ? '@' : ' @';
        $data['user_mentions'] = str_replace(' AND ', ' AND @', $data['user_mentions']);
        $data['user_mentions'] = str_replace(' OR ', ' OR @', $data['user_mentions']);
        $params['q'] .= $data['user_mentions'];
      }
      if ($data['hashtags'] != '') {
        $params['q'] .= ($params['q'] == '') ? '#' : ' #';
        $data['hashtags'] = str_replace(' AND ', ' AND #', $data['hashtags']);
        $data['hashtags'] = str_replace(' OR ', ' OR #', $data['hashtags']);
        $params['q'] .= $data['hashtags'];
      }
      if (!$data['include_retweets']) {
        $params['q'] .= ($params['q'] == '') ? '' : ' ';
        $params['q'] .= '-filter:retweets';
      }
      $endpoint = 'https://api.twitter.com/' . $api_version . '/search/tweets.json';
      break;

    case ST_SELECTION_MODE_FAVORITES:
      $params += array(
        'count' => 200,
      );
      $endpoint = 'https://api.twitter.com/' . $api_version . '/favorites/list.json';
      break;

    case ST_SELECTION_MODE_LIVE:
      $params += array(
        'count' => 200,
      );
      $endpoint = 'https://api.twitter.com/' . $api_version . '/statuses/user_timeline.json';
  }

  // If we don't have to load extra tweets, we might just as well request only
  // one tweet.
  if ($block_config['data']['count'] == 0) {
    $params['count'] = $block_config['data']['initial_count'];
  }

  // If we already cached tweets, only get tweets that come after the last
  // cached tweet.
  if ($since_id) {
    $params += array('since_id' => $since_id);
  }

  // Make the actual request to Twitter.
  $response = json_decode(
    $connection->oAuthRequest(
      $endpoint,
      'GET',
      $params
    )
  );

  $tweet_ids = array();

  if (is_object($response)) {
    if (isset($response->errors)) {
      foreach ($response->errors as $error) {
        watchdog('selective_tweets', 'Twitter API: @error_code: @error_message', array(
          '@error_code' => $error->code,
          '@error_message' => $error->message,
        ), WATCHDOG_ERROR);
      }
      throw new Exception(t('The tweets could not be loaded.'));
    }
    $response = $response->statuses;
  }

  foreach ($response as $tweet) {
    $tweet_ids[] = $tweet->id_str;
  }

  // Concatenate old and new tweets.
  $new_tweet_stack = array_merge($tweet_ids, $cached_tweet_ids);

  // Save the new tweet stack in cache.
  cache_set($cid, $new_tweet_stack, 'cache', CACHE_PERMANENT);

  // Update the cache_expire time for this Selective Tweets block.
  db_update('selective_tweets_block')
    ->fields(array(
      'cache_expire' => (REQUEST_TIME + ($block_config['data']['cached_time'] * 60)),
    ))
    ->condition('bid', $block_config['bid'])
    ->execute();

  return $new_tweet_stack;
}

/**
 * Helper function to convert hashtags, users and urls to links.
 *
 * @param $tweet
 *   The tweet object returned from the Twitter API.
 */
function selective_tweets_insert_links(&$tweet) {
  $twitter_url = 'https://twitter.com';

  // Convert hashtags.
  if (isset($tweet->entities->hashtags) && !empty($tweet->entities->hashtags)) {
    foreach ($tweet->entities->hashtags as $key => $hashtag) {
      $link = l('#' . $hashtag->text, $twitter_url . '/hashtag/' . $hashtag->text, array(
        'attributes' => array('target' => '_blank'),
        'query' => array('src' => 'hash'),
      ));
      $tweet->text = str_replace('#' . $hashtag->text, $link, $tweet->text);
    }
  }

  // Convert user mentions.
  if (isset($tweet->entities->user_mentions) && !empty($tweet->entities->user_mentions)) {
    foreach ($tweet->entities->user_mentions as $key => $user) {
      $link = l('@' . $user->screen_name, $twitter_url . '/' . $user->screen_name, array(
        'attributes' => array('target' => '_blank'),
      ));
      $tweet->text = str_replace('@' . $user->screen_name, $link, $tweet->text);
    }
  }

  // Convert urls.
  if (isset($tweet->entities->urls) && !empty($tweet->entities->urls)) {
    foreach ($tweet->entities->urls as $key => $url) {
      $link = l($url->display_url, $url->expanded_url, array(
        'attributes' => array('target' => '_blank'),
      ));
      $tweet->text = str_replace($url->url, $link, $tweet->text);
    }
  }
}
